
/*
 * xanim_ipc.c
 *
 * Copyright (C) 1995 by Mark Podlipec. 
 * All rights reserved.
 *
 * This software may be freely copied, modified and redistributed without
 * fee for non-commerical purposes provided that this copyright notice is
 * preserved intact on all copies and modified copies.
 * 
 * There is no warranty or other guarantee of fitness of this software.
 * It is provided solely "as is". The author(s) disclaim(s) all
 * responsibility and liability with respect to this software's usage
 * or its effect upon hardware or computer systems.
 *
 */
/****************
 * Rev History
 *
 * 03Jun95 - Created
 *
 *******************************/

#include "xanim.h"

#ifdef XA_FORK

#include "xanim_ipc.h"

static ULONG audio_debug_flag = FALSE;

#define AUD_DEBUG if (xa_debug == TRUE)

ULONG xa_audio_present;
ULONG xa_audio_status;

extern ULONG xa_audio_hard_buff;   /* AUDIO DOMAIN */
extern LONG xa_av_time_off;
extern void New_Merged_Audio_Output();

/**** Extern xanim_audio Functions *****/
extern void XA_Audio_Setup();
extern ULONG (*XA_Audio_Init)();
extern void (*XA_Audio_Kill)();
extern void (*XA_Audio_Off)();
extern void (*XA_Audio_On)();
extern ULONG (*XA_Closest_Freq)();
extern void  (*XA_Set_Output_Port)();
extern void  (*XA_Speaker_Tog)();
extern void  (*XA_Headphone_Tog)();
extern void  (*XA_LineOut_Tog)();
void XA_Audio_Init_Snd();


XA_AUD_FLAGS *audiof;

XA_AUD_HDR *xa_aud_hdr_start,*xa_aud_hdr_cur;
ULONG xa_aud_hdr_num = 0;

void XA_Audio_Child_Loop();
void XA_Child_Dies();
void XA_Child_Dies1();


ULONG XA_IPC_Receive();
ULONG XA_IPC_Send();
void XA_Audio_Child();
ULONG XA_Video_Send2_Audio();
ULONG XA_Audio_Receive_Video_CMD();
UBYTE *XA_Audio_Receive_Video_Buf();
ULONG XA_Audio_Send_ACK();
ULONG XA_Child_Find_File();
void XA_IPC_Reset_AV_Time();

extern XA_SND *xa_snd_cur;
extern LONG xa_time_audio;
extern ULONG xa_timelo_audio;

static ULONG xa_ipc_cmd_id = 1;

int xa_audio_fd[2];    /* audio child reads this, video writes this */
int xa_video_fd[2];    /* video reads this, audio child writes this */

LONG xa_audio_child = -1;

/************ XAnim Audio Child Code ***********************************/

/***************************************
 *  Routine for sending a buffer across the pipe. Handles partial writes
 *  caused to interrupts, full buffers, etc.
 *
 *  ? Implement Timeout ?
 **************/
ULONG XA_IPC_Send(fd,p,len,who)
int fd;
char *p;
int len;
int who;
{ while(len > 0)
  { int ret;
    ret = write( fd, p, len );
    if (ret < 0) 
    { AUD_DEBUG fprintf(stderr,"IPC(%ld) Send ERR: %ld\n",who,errno); 
      return(XA_IPC_ERR);
    }
    else { len -= ret; p += ret; }
  }
  return(XA_IPC_OK);
}

/***************************************
 *  Routine for receiving a buffer across the pipe. Handles partial reads
 *  caused to interrupts, empty buffers, etc.
 *
 *  ? Implement Select ?
 **************/
ULONG XA_IPC_Receive(fd,p,len,timeout,who)
int fd;
char *p;
int len;
LONG timeout;
int who;
{ LONG cur_time,ack_time;

  cur_time = ack_time = XA_Time_Read();
  ack_time += timeout;
  do
  { int ret;
    ret = read( fd, p, len);
    if (ret < 0) 
    { AUD_DEBUG fprintf(stderr,"IPC(%ld) Receive ERR %ld\n",who,errno); 
      return(XA_IPC_ERR);
    }
    else  /* ?POD overrun ever possible??? */
    { len -= ret; if (len <= 0) break;
      p += ret;
    }
    cur_time = XA_Time_Read();
  } while(cur_time < ack_time);

  if (len != 0)
  { AUD_DEBUG fprintf(stderr,"IPC(%ld) Receive TimeOut %ldms\n",who,timeout);
    return(XA_IPC_TOD);
  }

  return(XA_IPC_OK);
}

/***************************************
 * This routine blocks until something is ready on fd or
 * until a timeout occurrs.
 * 
 * Timeout is in milliseconds. 
 **************/
ULONG XA_IPC_Select(fd,timeout,who)
int fd;
LONG timeout;
int who;
{ int ret,tt_sec,tt_usec;
  int width = 64;
  fd_set readfds, writefds, exceptfds;
  struct timeval t_timeout;
 
  tt_sec = timeout / 1000;
  tt_usec = timeout - (tt_sec * 1000);
  tt_usec *= 1000;
  
/* AUD_DEBUG fprintf(stderr,"tt_sec %ld tt_usec %ld\n",tt_sec,tt_usec); */

  FD_ZERO(&readfds);
  FD_SET( fd , &readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  t_timeout.tv_sec =  tt_sec;
  t_timeout.tv_usec = tt_usec;
 
  ret = select(width, &readfds, &writefds, &exceptfds, &t_timeout);
 
  if (ret < 0) AUD_DEBUG fprintf(stderr,"AUD select ERR: %ld\n",errno);
  return(ret);
}

/***************************************
 *  Routine for having the Video Process send a Command with
 *  optional Command Acknowledgement and option data buffer.
 *
 *  ack_flag is number of ms to wait for ACK. If 0, then no ACK is checked
 *  for.
 **************/
ULONG XA_Video_Send2_Audio(cmd,buf,buf_len,value,ack_flag,ack_val)
ULONG cmd;
UBYTE *buf;
ULONG buf_len;
ULONG value;
ULONG ack_flag;
ULONG *ack_val;
{ XA_IPC_HDR ipc_cmd,ipc_ack;
  ULONG ret;
  LONG ack_time, cur_time, len;
  char *p;

  /*** Setup Command */
  ipc_cmd.cmd = cmd;
  ipc_cmd.time = XA_Time_Read();
  ipc_cmd.len = buf_len;
  ipc_cmd.value = value;
  ipc_cmd.id  = xa_ipc_cmd_id;   
  xa_ipc_cmd_id++;

  /*** SEND IPC Command */
  ret = XA_IPC_Send( xa_audio_fd[XA_FD_WRITE], ((char *)(&ipc_cmd)), 
					(sizeof(XA_IPC_HDR)), XA_IAM_VIDEO );
  if (ret == XA_IPC_ERR) 
	{ AUD_DEBUG fprintf(stderr,"Vid Send Cmd Err\n"); return(FALSE); }


  /* SEND Data if Any */
  p = (char *)(buf);
  len = buf_len;
  while(len > 0)
  { int sel_ret,tlen = len;

    if (tlen > XA_IPC_CHUNK) tlen = XA_IPC_CHUNK;
    AUD_DEBUG fprintf(stderr,"VID IPC Sendin Chunk tlen %ld\n",tlen);
    ret = XA_IPC_Send( xa_audio_fd[XA_FD_WRITE], p, tlen, XA_IAM_VIDEO );
    if (ret == XA_IPC_ERR)
	{ AUD_DEBUG fprintf(stderr,"Vid Send Buf Err\n"); return(FALSE); }
    p += tlen;;  len -= tlen;

    /*** wait for ACK */
    sel_ret = XA_IPC_Select(xa_video_fd[XA_FD_READ], ack_flag,XA_IAM_VIDEO); 
    if (sel_ret <= 0)
    {
      AUD_DEBUG fprintf(stderr,"VID: chunk Ack Timeout/err: cmd %lx %ld\n",cmd,sel_ret);
      return(FALSE);
    }

    /*** Read ACK */
    ret = XA_IPC_Receive(xa_video_fd[XA_FD_READ],((char *)(&ipc_ack)),
		(sizeof(XA_IPC_HDR)),ack_flag,XA_IAM_VIDEO);
    if (ret != XA_IPC_OK)
	{ AUD_DEBUG fprintf(stderr,"Vid chunk Ack Err\n"); return(FALSE); }
    AUD_DEBUG fprintf(stderr,"VID IPC Sent Chunk OK\n");
  }

  /*** Look For ACK */
  if (ack_flag)
  { int sel_ret = XA_IPC_Select(xa_video_fd[XA_FD_READ], ack_flag,XA_IAM_VIDEO);
    if (sel_ret <= 0)
    {
      AUD_DEBUG fprintf(stderr,"VID: Ack Timeout/err: cmd %lx %ld\n",cmd,sel_ret);
      return(FALSE);
    }

    ret = XA_IPC_Receive(xa_video_fd[XA_FD_READ],((char *)(&ipc_ack)),
		(sizeof(XA_IPC_HDR)),ack_flag,XA_IAM_VIDEO);
    if (ret == XA_IPC_ERR) 
		{AUD_DEBUG fprintf(stderr,"Vid IPC Ack Err\n"); return(FALSE);}
    else if (ret == XA_IPC_TOD) 
		{AUD_DEBUG fprintf(stderr,"Vid IPC Ack Timeout\n"); return(FALSE);}
    if (ipc_cmd.id != ipc_ack.id) 
    {
      AUD_DEBUG fprintf(stderr,"VID IPC ID mismatch %ld %ld cmd %lx\n",
					ipc_cmd.id,ipc_ack.id,ipc_cmd.cmd);
      return(FALSE); 
    }
    if (ack_val) *ack_val = ipc_ack.value;
  }

  AUD_DEBUG fprintf(stderr,"VID IPC Success %ld %ld cmd %lx\n",
	ipc_cmd.time,ipc_ack.time,ipc_cmd.cmd);
  return(TRUE);
}

/***************************************
 * Accept Command from Video Process.  Returns CMD on success 
 * or XA_IPC_ERR on failure.
 * 
 **************/
ULONG XA_Audio_Receive_Video_CMD(ipc_cmd,timeout)
XA_IPC_HDR *ipc_cmd;
LONG timeout;
{ ULONG ret;

  ret = XA_IPC_Receive(xa_audio_fd[XA_FD_READ], ((char *)(ipc_cmd)),
				(sizeof(XA_IPC_HDR)),timeout, XA_IAM_AUDIO );
  if (ret == XA_IPC_ERR) 
	{ AUD_DEBUG fprintf(stderr,"AUD Receive CMD Err\n"); return(XA_IPC_ERR); }
  else if (ret == XA_IPC_TOD)
	{ AUD_DEBUG fprintf(stderr,"AUD Receive CMD TOD\n"); return(XA_IPC_TOD); }
  
  AUD_DEBUG fprintf(stderr,"AUD SUCCESS! cmd %lx len %ld id %ld\n",
			ipc_cmd->cmd,ipc_cmd->len,ipc_cmd->id);
  return(XA_IPC_OK);
}


/***************************************
 * Accept Buffer from Video Process.  Returns buff_len on success 
 * or 0 on failure.
 * 
 **************/
UBYTE *XA_Audio_Receive_Video_Buf(len,timeout)
ULONG len;	/* len of buffer being sent (from ipc_cmd.len) */
LONG timeout;
{ int blen;
  UBYTE *b,*p;
  
  b = (UBYTE *)malloc( len );
  if (b == 0) {  AUD_DEBUG fprintf(stderr,"AUD Rx BUF: malloc err\n"); return(0); }

  p = b;
  blen = len;
  while(blen > 0)
  { int ret,tlen;
    ret = XA_IPC_Select(xa_audio_fd[XA_FD_READ], 500,XA_IAM_AUDIO);
    if (ret <= 0) { free(b); return(0); }

    tlen = blen; if (tlen > XA_IPC_CHUNK) tlen = XA_IPC_CHUNK;
    ret = read( xa_audio_fd[XA_FD_READ], p, tlen);
    if (ret != tlen)  /* POD improve: make while() etc */
	{ AUD_DEBUG fprintf(stderr,"read err %ld %ld\n",ret,tlen); free(b); return(0); }
    p += tlen; 
    blen -= tlen;
    XA_Audio_Send_ACK(XA_IPC_OK,0,0);
  }
  return(b);
}


/***************************************
 * This routines sends an Acknowledgement back the Video Process.
 * 
 **************/
ULONG XA_Audio_Send_ACK(ack,id,value)
ULONG ack;
{ XA_IPC_HDR ipc_ack;
  ULONG ret;
  LONG len;
  char *p;

  /****************** Send ACK Back to Video */
  ipc_ack.cmd = ack;
  ipc_ack.time = XA_Time_Read();
  ipc_ack.len = 0;
  ipc_ack.id  = id;
  ipc_ack.value = value;
  ret = XA_IPC_Send( xa_video_fd[XA_FD_WRITE], ((char *)(&ipc_ack)), 
					(sizeof(XA_IPC_HDR)), XA_IAM_AUDIO );
  if (ret == XA_IPC_ERR) 
	{ AUD_DEBUG fprintf(stderr,"AUD Send ACK Err\n"); return(XA_IPC_ERR); }
  return(ack);
}


/***************************************
 * This routines cleans up after the Child and then
 * exits.
 *
 **************/
void XA_Child_Dies()
{ XA_AUD_HDR *aud_hdr;


AUD_DEBUG fprintf(stderr,"CHILD IS DYING\n");
  aud_hdr = xa_aud_hdr_start;
  if (aud_hdr) xa_aud_hdr_start->prev->next = 0;  /* break loop */
  while(aud_hdr)
  { XA_AUD_HDR *tmp_hdr = aud_hdr->next;
   /* MYFREE (aud_hdr->snd) loop */
   if (aud_hdr->filename) free(aud_hdr->filename);
   free(aud_hdr);
   aud_hdr = tmp_hdr;
  }
  if (audiof)
  {
    if (audiof->device) free(audiof->device);
    free(audiof);
  }
AUD_DEBUG fprintf(stderr,"CHILD IS DEAD\n");
  exit(0);
}

/***************************************
 * This routine prints out a message and then calls XA_Child_Dies.
 *
 **************/
void XA_Child_Dies1(s)
char *s;
{
  AUD_DEBUG fprintf(stderr,"CHILD: %s\n",s);
  XA_Child_Dies();
}

/***************************************
 * This routine returns a XA_AUD_HDR structure.
 * and removes it from the loop.
 *
 **************/
XA_AUD_HDR *Return_Aud_Hdr(aud_hdr)
XA_AUD_HDR *aud_hdr;
{ XA_AUD_HDR *tmp_hdr;
  if ((aud_hdr==0) || (xa_aud_hdr_start==0)) TheEnd1("Return_Anim_Hdr err");
  xa_aud_hdr_num--;
  if (xa_aud_hdr_num == 0)
  {
    xa_aud_hdr_start = 0;
    tmp_hdr = 0;
  }
  else /* removed aud_hdr from the loop */
  {
    tmp_hdr		= aud_hdr->prev;
    tmp_hdr->next	= aud_hdr->next;
    aud_hdr->next->prev	= tmp_hdr;
  }
  if (aud_hdr->filename) free(aud_hdr->filename);
  free(aud_hdr);
  return(tmp_hdr);
}


/***************************************
 * This routine allocates a XA_AUD_HDR structure.
 *  aud_file is assumed to be consumable.
 *
 **************/
XA_AUD_HDR *Get_Aud_Hdr(aud_hdr,num,aud_file)
XA_AUD_HDR *aud_hdr;
ULONG num;
char *aud_file;
{
  XA_AUD_HDR *temp_hdr;
  temp_hdr = (XA_AUD_HDR *)malloc( sizeof(XA_AUD_HDR) );
  if (temp_hdr == 0) XA_Child_Dies1("Get_Anim_Hdr: malloc failed\n");

  temp_hdr->num = num;
  temp_hdr->filename = aud_file;
  temp_hdr->first_snd = 0;
  temp_hdr->last_snd = 0;

  if (aud_hdr == 0)
  {
    xa_aud_hdr_start  = temp_hdr;
    temp_hdr->next = temp_hdr;
    temp_hdr->prev = temp_hdr;
  }
  else
  {
    temp_hdr->prev   = aud_hdr;
    temp_hdr->next   = aud_hdr->next;
    aud_hdr->next    = temp_hdr;
    xa_aud_hdr_start->prev = temp_hdr;
  }
  return(temp_hdr);
}

/***************************************
 *
 *
 **************/
void XA_Audio_Child()
{
  xa_aud_hdr_start = xa_aud_hdr_cur = 0;
  xa_aud_hdr_num = 0;

  audiof = (XA_AUD_FLAGS *)malloc( sizeof(XA_AUD_FLAGS) );
  if (audiof==0) XA_Child_Dies1("audiof malloc err");

/* don't init */
  audiof->enable	= FALSE;
  audiof->mute		= FALSE;
  audiof->newvol	= TRUE;
  audiof->divtest	= 2;
  audiof->fromfile	= FALSE;
  audiof->bufferit	= FALSE;

  audiof->port		= DEFAULT_XA_AUDIO_PORT;
  audiof->volume	= 0;
  audiof->playrate	= 0;
  audiof->device	= 0;


/* POD DEBUGGING PURPOSES */

AUD_DEBUG fprintf(stderr,"CHILD IS AWAKE\n");
  XA_Audio_Child_Loop();
AUD_DEBUG fprintf(stderr,"CHILD IS DONE\n");

  XA_Child_Dies();
}


/***************************************
 *
 *
 **************/
void XA_Audio_Child_Loop()
{ ULONG exitflag;
  int flag = 0;


  exitflag = FALSE;
  while( exitflag != TRUE )
  { int ret;
    XA_IPC_HDR ipc_cmd;
    UBYTE *ipc_buff;

    /* Wait for Incoming */
    ret = XA_IPC_Select(xa_audio_fd[XA_FD_READ], 10,XA_IAM_AUDIO);
    if (ret < 0) return;		/* error occurred - get out */
    else if (ret == 0) 
    {
      if (xa_audio_status == XA_AUDIO_STARTED) New_Merged_Audio_Output();
      continue;	/* nothing ready - loop again */
    }
    /* else fall through - something is ready */

    XA_Audio_Receive_Video_CMD(&ipc_cmd, 5000);

    switch( ipc_cmd.cmd )
    {
	/* send back status = STOPPED */
      case XA_IPC_AUD_SETUP:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_SETUP\n",ipc_cmd.cmd);
	XA_Audio_Setup();
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_status);
	break;

      case XA_IPC_AUD_INIT:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_INIT\n",ipc_cmd.cmd);
	if (xa_audio_present == XA_AUDIO_UNK) XA_Audio_Init();
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_present);
	break;

      case XA_IPC_AUD_KILL:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_KILL\n",ipc_cmd.cmd);
	XA_Audio_Kill();
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_status);
	break;

	/* solely for returning audio status */
      case XA_IPC_GET_STATUS:
	AUD_DEBUG fprintf(stderr,"AUD IPC: GET_STATUS\n",ipc_cmd.cmd);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_status);
	break;

      case XA_IPC_GET_PRESENT:
	AUD_DEBUG fprintf(stderr,"AUD IPC: GET_PRESENT\n",ipc_cmd.cmd);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_present);
	break;

      case XA_IPC_AUD_ON:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_ON  %lx\n",xa_snd_cur);
	XA_Audio_On();
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_status);
	break;

      case XA_IPC_AUD_OFF:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_OFF\n",ipc_cmd.cmd);
	XA_Audio_Off(ipc_cmd.value);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_status);
	break;

      case XA_IPC_AUD_PORT:
	audiof->port = ipc_cmd.value;
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_PORT\n",ipc_cmd.cmd);
	XA_Set_Output_Port(ipc_cmd.value);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_STOG:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_STOG\n",ipc_cmd.cmd);
	XA_Speaker_Tog(ipc_cmd.value);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_HTOG:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_HTOG\n",ipc_cmd.cmd);
	XA_Headphone_Tog(ipc_cmd.value);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_LTOG:
	AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_LTOG\n",ipc_cmd.cmd);
	XA_LineOut_Tog(ipc_cmd.value);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_GET_CFREQ:
	{ ULONG hfreq = ipc_cmd.value;
	  AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_CFREQ %lx freq %ld\n",(ULONG)(XA_Closest_Freq),hfreq);
	  hfreq = XA_Closest_Freq(hfreq);
AUD_DEBUG fprintf(stderr,"CFREQ: hfreq = %ld\n",hfreq);
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,hfreq);
	}
	break;

/*POD NOTE: CFREQ must be called before this */
      case XA_IPC_GET_BSIZE:
	{ AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_BSIZE\n",ipc_cmd.cmd);
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,xa_audio_hard_buff);
	}
	break;

      case XA_IPC_FILE:
	AUD_DEBUG fprintf(stderr,"AUD IPC: received FILE\n",ipc_cmd.cmd);
        if (ipc_cmd.len) 
	{
	  ipc_buff =  XA_Audio_Receive_Video_Buf(ipc_cmd.len,500);
	  if (ipc_buff)
	  { AUD_DEBUG fprintf(stderr,"FILE: %s \n",ipc_buff);
	    xa_aud_hdr_cur = Get_Aud_Hdr(xa_aud_hdr_cur,ipc_cmd.value,ipc_buff);
	    ipc_buff = 0;  /* indicate we're using it */
	  }
	  else { AUD_DEBUG fprintf(stderr,"FILE: err\n"); }
	  ipc_cmd.len = 0; /* indicate we've read it */
	} 
	XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_N_FILE:
	{ ULONG ok;
	  AUD_DEBUG fprintf(stderr,"AUD IPC: received N_FILE\n",ipc_cmd.cmd);
	  if (xa_audio_present==XA_AUDIO_OK)
			ok = XA_Child_Find_File(ipc_cmd.value,0);
	  else ok = FALSE;
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,ok);
	}
	break;

      case XA_IPC_P_FILE:
	{ ULONG ok;
	  AUD_DEBUG fprintf(stderr,"AUD IPC: received P_FILE\n",ipc_cmd.cmd);
	  if (xa_audio_present==XA_AUDIO_OK)
			ok = XA_Child_Find_File(ipc_cmd.value,1);
	  else ok = FALSE;
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,ok);
	}
	break;

      case XA_IPC_SND_INIT:
	{ 
	  AUD_DEBUG fprintf(stderr,"AUD IPC: SND_INIT\n",ipc_cmd.cmd);
	  XA_Audio_Init_Snd(xa_snd_cur);
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	}
	break;

      case XA_IPC_SND_ADD:
	{ XA_SND *new_snd = 0;
	  AUD_DEBUG fprintf(stderr,"AUD IPC: SND_ADD\n",ipc_cmd.cmd);
	  if (ipc_cmd.len)
	  {
	    ULONG sret;
            new_snd = (XA_SND *)
			XA_Audio_Receive_Video_Buf(ipc_cmd.len,500);
            if (new_snd)	sret = XA_IPC_Sound(xa_aud_hdr_cur,new_snd);
	    else		sret = 0;
            ipc_cmd.len = 0;
            XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,sret);
	  } 
	  else XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,FALSE);

          /* SND_BUF MUST follow */
          XA_Audio_Receive_Video_CMD(&ipc_cmd, 5000);
          if (ipc_cmd.cmd != XA_IPC_SND_BUF)
		XA_Child_Dies1("SND_BUF Did NOT follow SND_ADD\n");
          AUD_DEBUG fprintf(stderr,"AUD IPC: SND_BUF\n",ipc_cmd.cmd);
          if (ipc_cmd.len)
          { ULONG sret = FALSE;
	    if (new_snd) new_snd->snd = 
		(UBYTE *)XA_Audio_Receive_Video_Buf(ipc_cmd.len,500);
	    if (new_snd->snd) sret = TRUE;
            XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,sret);
            ipc_cmd.len = 0;
          }
          else
	  {
	    new_snd->snd = 0;
            XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,TRUE);
	  }
	}
	break;
      case XA_IPC_VID_TIME:
	AUD_DEBUG fprintf(stderr,"AUD IPC: VID_TIME\n",ipc_cmd.cmd);
	xa_av_time_off = ipc_cmd.value;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_RST_TIME:
	{ ULONG tt = XA_Time_Read();
	  AUD_DEBUG fprintf(stderr,"AUD IPC: RST_TIME time %ld\n",tt);
          XA_IPC_Reset_AV_Time(ipc_cmd.value);
	  tt = XA_Time_Read();
	  AUD_DEBUG fprintf(stderr,"AUD IPC: time end %ld\n",tt);
          XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	}
	break;

      case XA_IPC_AUD_ENABLE:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_ENABLE\n",ipc_cmd.cmd);
	audiof->enable = ipc_cmd.value;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_MUTE:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_MUTE\n",ipc_cmd.cmd);
	audiof->mute = ipc_cmd.value;
	audiof->newvol = TRUE;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_VOL:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_VOL\n",ipc_cmd.cmd);
	audiof->volume = ipc_cmd.value;
	audiof->newvol = TRUE;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_RATE:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_RATE\n",ipc_cmd.cmd);
	audiof->playrate = ipc_cmd.value;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_DEV:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_DEV\n",ipc_cmd.cmd);
        if (ipc_cmd.len) 
	{
	  ipc_buff =  XA_Audio_Receive_Video_Buf(ipc_cmd.len,500);
	  if (ipc_buff)
	  { AUD_DEBUG fprintf(stderr,"DEVICE: %s \n",ipc_buff);
	    audiof->device = (char *)(ipc_buff);
	    ipc_buff = 0;
	  }
	  else { AUD_DEBUG fprintf(stderr,"FILE: err\n"); }
	  ipc_cmd.len = 0; /* indicate we've read it */
	}
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_FFLAG:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_FFLAG\n",ipc_cmd.cmd);
	audiof->fromfile = ipc_cmd.value;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_AUD_BFLAG:
        AUD_DEBUG fprintf(stderr,"AUD IPC: AUD_BFLAG\n",ipc_cmd.cmd);
	audiof->bufferit = ipc_cmd.value;
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_ERR:
        XA_Audio_Send_ACK(XA_IPC_ACK_ERR,ipc_cmd.id,0);
	return;
	break;

      case XA_IPC_HELLO:
	AUD_DEBUG fprintf(stderr,"AUD IPC: received Hello\n",ipc_cmd.cmd);
        XA_Audio_Send_ACK(XA_IPC_ACK_OK,ipc_cmd.id,0);
	break;

      case XA_IPC_EXIT:
	AUD_DEBUG fprintf(stderr,"AUD IPC: received EXIT\n",ipc_cmd.cmd);
        XA_Audio_Send_ACK(XA_IPC_ACK_BYE,ipc_cmd.id,0);
	return;
	break;

      default:
	AUD_DEBUG fprintf(stderr,"AUD IPC: unknown cmd %ld\n",ipc_cmd.cmd);
        XA_Audio_Send_ACK(XA_IPC_ACK_ERR,ipc_cmd.id,0);
	return;
	break;
    }
    /* Flush len if no command uses it */
    if (ipc_cmd.len)
    {
	ipc_buff =  XA_Audio_Receive_Video_Buf(ipc_cmd.len,500);
        free(ipc_buff); ipc_buff = 0;
        XA_Audio_Send_ACK(XA_IPC_ACK_ERR,ipc_cmd.id,0);
        return;
    } else ipc_buff = 0;

  } 
}


/*********************************
 *  This routine sets up the xa_audio_fd and xa_video_fd pipes, then
 *  forks off the Audio Child process.
 *
 *   + The Parent returns(TRUE) on success, FALSE on failure.
 *   + The Child calls XA_Audio_Child().
 *
 ****************/
ULONG XA_Give_Birth()
{ int ret;

  ret = pipe( xa_audio_fd );
  if (ret) { AUD_DEBUG fprintf(stderr,"PIPE for audio failed: %ld\n",errno); }
  else
  {
    ret = pipe( xa_video_fd );
    if (ret) { AUD_DEBUG fprintf(stderr,"PIPE for video failed: %ld\n",errno); }
    else
    {
      xa_audio_child = fork();
      if (xa_audio_child == 0)   /* I am the Audio Child */
      { 
        XA_Audio_Child(); 
        exit(0);
      }
      else if (xa_audio_child < 0) /* Still Born */
      {
	AUD_DEBUG fprintf(stderr,"Audio Child Still Born: %ld\n",errno);
      }
      else			/* I am the Video Parent */
      { 
        return(TRUE); 
      }
    }
  }
  return(FALSE);
}


/*********************************
 * Move along Audio Header til num is matched. 
 * returns TRUE if found AND has valid snd.
 * Else returns FALSE.
 *
 * Set's global variable xa_snd_cur
 ****************/
ULONG XA_Child_Find_File(num,flag)
ULONG num;
ULONG flag;	/* 0 means next, 1 means prev */
{ XA_AUD_HDR *cur = xa_aud_hdr_cur;
  do
  {
    if (cur->num == num) 
    { 
      xa_aud_hdr_cur = cur; 
      xa_snd_cur = cur->first_snd;
      if (xa_snd_cur)		return(TRUE);
      else			return(FALSE);
    }
    cur = (flag == 0)?(cur->next):(cur->prev);
  } while(cur != xa_aud_hdr_cur);
}

/*********************************
 *
 ****************/
void XA_IPC_Reset_AV_Time(vid_time)
LONG vid_time;
{ int xflag = FALSE;

    if (xa_snd_cur==0) { AUD_DEBUG fprintf(stderr,"AA\n"); return; }
    XA_Audio_Init_Snd(xa_snd_cur);
    /* Move to the correct snd chunk */
    while(xflag == FALSE)
    { LONG snd_time = xa_snd_cur->snd_time;
      if (snd_time > vid_time)
      { XA_SND *p_snd = xa_snd_cur->prev;
        AUD_DEBUG fprintf(stderr,"s>v %ld %ld\n",snd_time,vid_time);
        if (p_snd) xa_snd_cur = p_snd;
        else xflag = TRUE;
      }
      else if (snd_time < vid_time)
      { XA_SND *n_snd = xa_snd_cur->next;
        AUD_DEBUG fprintf(stderr,"s<v %ld %ld\n",snd_time,vid_time);
        if (n_snd) 
        {
	  if (n_snd->snd_time <= vid_time) xa_snd_cur = n_snd;
	  else xflag = TRUE;
        }
        else xflag = TRUE;
      }
      else 
      {
        AUD_DEBUG fprintf(stderr,"s=v %ld %ld\n",snd_time,vid_time);
        xflag = TRUE;
      }
    } /* end while xflag */

    /* Move within the snd chunk - HAVE NOP FLAG */
    if (xa_snd_cur)
    { XA_SND *shdr = xa_snd_cur;
	/* read in from file if needed */
      if (xa_snd_cur->fpos >= 0)
      { 
/* POD TEMP
        xa_snd_cur->snd = xa_audcodec_buf;
        XA_Read_Audio_Delta(xa_aud_fd,xa_snd_cur->fpos,
				xa_snd_cur->tot_bytes,xa_audcodec_buf);
*/
      } 
      { ULONG tmp_cnt; LONG diff;

	/* time diff in ms */
        diff =  (vid_time - shdr->snd_time); if (diff < 0) diff = 0;
	/* calc num of samples in that time frame */
	tmp_cnt = (diff * shdr->ifreq) / 1000; 
	if (tmp_cnt & 0x01) tmp_cnt--;  /* make multiple of 2 for ADPCM */
	/* Init snd_hdr */
        XA_Audio_Init_Snd(xa_snd_cur);
	if (tmp_cnt) /* not at beginning */
	{ char *garb; /* play sound into garb buffer */
	  garb = (char *)malloc(4 * tmp_cnt);
          if (garb) 
	  { diff = tmp_cnt - xa_snd_cur->delta(xa_snd_cur,garb,0,tmp_cnt);
	    free(garb);
	    if (diff != 0) fprintf(stderr,"AV Warn: rst sync err %lx\n",diff);
	  }
	}
      }
      xa_time_audio = vid_time;
      xa_timelo_audio = 0;
    } /* end of valid xa_snd_cur */ 
}

void XA_IPC_Close_Pipes()
{
  if (xa_audio_fd[0] >= 0) { close(xa_audio_fd[0]); xa_audio_fd[0] = -1; }
  if (xa_audio_fd[1] >= 0) { close(xa_audio_fd[1]); xa_audio_fd[1] = -1; }
  if (xa_video_fd[0] >= 0) { close(xa_video_fd[0]); xa_video_fd[0] = -1; }
  if (xa_video_fd[1] >= 0) { close(xa_video_fd[1]); xa_video_fd[1] = -1; }
}

void XA_IPC_Set_Debug(value)
ULONG value;
{
  audio_debug_flag = value;
}

#endif
